# Generated by Django 5.0.2 on 2024-08-21 15:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hostore', '0006_testdatastorewithattribute'),
    ]

    operations = [
        migrations.CreateModel(
            name='TestTimeseriesStoreWithAttribute',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('data', models.BinaryField(blank=True, null=True)),
                ('version', models.IntegerField(default=0)),
                ('year', models.IntegerField()),
                ('kind', models.CharField(max_length=100)),
            ],
            options={
                'abstract': False,
                'indexes': [models.Index(fields=['version'], name='hostore_tes_version_822f4b_idx')],
                'unique_together': {('version', 'year', 'kind')},
            },
        ),
    ]
