# Generated by Django 5.0.2 on 2024-05-29 08:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hostore', '0005_alter_testdatastore_prm'),
    ]

    operations = [
        migrations.CreateModel(
            name='TestDataStoreWithAttribute',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('client_id', models.IntegerField()),
                ('prm', models.CharField(max_length=70)),
                ('last_modified', models.DateTimeField(auto_now=True)),
                ('created_at', models.DateTimeField(auto_now_add=True, null=True)),
                ('data', models.BinaryField(blank=True, null=True)),
                ('version', models.IntegerField(default=0)),
                ('year', models.IntegerField()),
            ],
            options={
                'abstract': False,
                'indexes': [models.Index(fields=['prm', 'client_id'], name='hostore_tes_prm_93d669_idx')],
                'unique_together': {('prm', 'client_id', 'year', 'created_at')},
            },
        ),
    ]
