# Generated by Django 5.2.4 on 2025-07-18 14:22

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('hostore', '0008_remove_testtimeseriesstorewithattribute_hostore_tes_version_822f4b_idx_and_more'),
    ]

    operations = [
        migrations.DeleteModel(
            name='TestDataStore',
        ),
        migrations.DeleteModel(
            name='TestDataStoreWithAttribute',
        ),
        migrations.DeleteModel(
            name='TestTimeseriesStoreWithAttribute',
        ),
    ]
