# Generated by Django 5.0.2 on 2024-08-21 16:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hostore', '0007_testtimeseriesstorewithattribute'),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name='testtimeseriesstorewithattribute',
            name='hostore_tes_version_822f4b_idx',
        ),
        migrations.AlterUniqueTogether(
            name='testtimeseriesstorewithattribute',
            unique_together={('year', 'kind')},
        ),
        migrations.AddIndex(
            model_name='testtimeseriesstorewithattribute',
            index=models.Index(fields=['data'], name='hostore_tes_data_9fd074_idx'),
        ),
        migrations.RemoveField(
            model_name='testtimeseriesstorewithattribute',
            name='version',
        ),
    ]
